local _G = _G

local Cartographer3 = _G.Cartographer3

Cartographer3.version = GetAddOnMetadata("Cartographer3", "Version")
Cartographer3.date = nil
do
	local timestamp = tonumber(GetAddOnMetadata("Cartographer3", "X-Timestamp"))
	if timestamp then
		Cartographer3.date = date("%B %d, %Y, %H:%M:%S", timestamp)
	end
end
do
	local rawVersion = Cartographer3.version:match("^3.0 v(.*)$")
	if not rawVersion then
		Cartographer3.versionType = "Development"
		Cartographer3.version = "3.0 Development"
	elseif rawVersion:match("%-%d+%-g%x+") then
		Cartographer3.versionType = "Alpha"
	elseif not rawVersion:match("^[%d%-_%.]+$") then
		Cartographer3.versionType = "Beta"
	else
		Cartographer3.versionType = "Release"
	end
end

local db

local L = _G.Cartographer3__L
_G.Cartographer3__L = nil
Cartographer3.L = L

local Cartographer3_Utils = Cartographer3.Utils

local BLIZZARD_MAP_WIDTH = 1002
local BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO = 2/3
local BLIZZARD_MINIMAP_TILE_YARD_SIZE = 533 + 1/3

local Cartographer3_Data = Cartographer3.Data
-- TODO: Separate Data from Constants
Cartographer3_Data.CONTINENT_DATA = {}
Cartographer3_Data.ZONE_DATA = {}
Cartographer3_Data.LOCALIZED_ZONE_TO_TEXTURE = {}
Cartographer3_Data.currentMapTexture = GetMapInfo()
Cartographer3_Data.currentContinentID = GetCurrentMapContinent()

function Cartographer3.SetConstants(data)
	local CONTINENT_DATA = Cartographer3_Data.CONTINENT_DATA
	Cartographer3.SetConstants = nil
	for k, v in pairs(data) do
		Cartographer3_Data[k] = v
	end
	for k, v in pairs(data.CONTINENT_DATA) do
		CONTINENT_DATA[k] = v
	end
	data.CONTINENT_DATA = CONTINENT_DATA
	
	-- minimap width in pixels / yards at closest radius * yards per pixel
	Cartographer3_Data.MAXIMUM_ZOOM = 140 / (133 + 1/3) * Cartographer3_Data.YARDS_PER_PIXEL
	
	-- able to see twice what the minimap can show
	if not Cartographer3_Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE then
		Cartographer3_Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE = Cartographer3_Data.MAXIMUM_ZOOM / 7
	end
end

function Cartographer3.SetOverlayData(data)
	Cartographer3.SetOverlayData = nil
	Cartographer3_Data.OVERLAY_DATA = data
end

function Cartographer3.SetTextureData(data)
	Cartographer3.SetTextureData = nil
	Cartographer3_Data.TEXTURE_DATA = data
end

function Cartographer3.SetInstanceTextureData(data)
	Cartographer3.SetInstanceTextureData = nil
	Cartographer3_Data.INSTANCE_TEXTURE_DATA = data
end

local mapHolder, mapFrame, scrollFrame, mapView

local WorldMapFrame = _G.WorldMapFrame
WorldMapTooltip:SetParent(UIParent)
WorldMapTooltip:SetFrameStrata("TOOLTIP")
local currentOwner, currentPoint
local old_WorldMapTooltip_SetOwner = WorldMapTooltip.SetOwner
function WorldMapTooltip.SetOwner(this, owner, point)
	-- do a SetOwner hack since if you set it to the owner inside the mapView, then there are scaling issues and it would be stuck inside the scroll child. This way, it gives the same effect as the author intended.
	currentOwner = owner
	currentPoint = point
	old_WorldMapTooltip_SetOwner(this, mapHolder, "ANCHOR_NONE")
	this:ClearAllPoints()
	if point == "ANCHOR_TOPRIGHT" then
		this:SetPoint("BOTTOMRIGHT", currentOwner, "TOPRIGHT")
	elseif point == "ANCHOR_RIGHT" then
		this:SetPoint("BOTTOMLEFT", currentOwner, "TOPRIGHT")
	elseif point == "ANCHOR_BOTTOMRIGHT" then
		this:SetPoint("TOPLEFT", currentOwner, "BOTTOMRIGHT")
	elseif point == "ANCHOR_TOPLEFT" then
		this:SetPoint("BOTTOMLEFT", currentOwner, "TOPLEFT")
	elseif point == "ANCHOR_LEFT" then
		this:SetPoint("BOTTOMRIGHT", currentOwner, "TOPLEFT")
	elseif point == "ANCHOR_BOTTOMLEFT" then
		this:SetPoint("TOPRIGHT", currentOwner, "BOTTOMLEFT")
	elseif point == "ANCHOR_CURSOR" then
		old_WorldMapTooltip_SetOwner(this, mapHolder, "ANCHOR_CURSOR")
	elseif point == "ANCHOR_PRESERVE" then
		this:SetPoint("BOTTOM", UIParent, "BOTTOM", 0, 80)
	end
end
function WorldMapTooltip.IsOwned(this, owner)
	return currentOwner == owner
end
local function HideWorldMapFrames(...)
	for i = 1, select('#', ...) do
		local v = select(i, ...)
		local name = v:GetName()
		if name and name:match("^WorldMap") then
			v:Hide()
			v:SetParent(nil)
		end
	end
end	
UIPanelWindows["WorldMapFrame"] = nil
WorldMapFrame:SetAttribute("UIPanelLayout-enabled", false)
WorldMapFrame:SetFrameStrata("MEDIUM")
WorldMapFrame:EnableMouse(nil)
WorldMapFrame:EnableMouseWheel(nil)
WorldMapFrame:ClearAllPoints()
WorldMapFrame:SetWidth(1024)
WorldMapFrame:SetHeight(768)
HideWorldMapFrames(_G.WorldMapDetailFrame:GetRegions())
HideWorldMapFrames(_G.WorldMapButton:GetChildren())
_G.WorldMapButton:EnableMouse(false)
_G.WorldMapButton:SetScript("OnUpdate", nil)
function _G.WorldMapButton_OnClick()
	-- hopefully I don't destroy another addon's hooks here, but *shrug*
end
_G.WorldMapButton:SetScript("OnEvent", nil)
_G.WorldMapButton:UnregisterAllEvents()
WorldMapFrame:SetScript("OnShow", nil)
WorldMapFrame:SetScript("OnHide", nil)
WorldMapFrame:SetScript("OnEvent", nil)
WorldMapFrame:UnregisterAllEvents()
WorldMapFrame:SetScript("OnUpdate", nil)
WorldMapFrame:SetScript("OnKeyDown", nil)
_G.WorldMapDetailFrame:ClearAllPoints()
_G.WorldMapButton:SetParent(_G.WorldMapDetailFrame)
WorldMapFrame:SetPoint("CENTER", _G.WorldMapDetailFrame, "CENTER", 2, 17) -- get relatively in the same position
function WorldMapFrame:GetScale()
	-- hack to make old CTMod code work
	return self:GetEffectiveScale()
end
_G.WorldMapPositioningGuide:SetPoint("CENTER", _G.WorldMapDetailFrame, "CENTER", 2, 17)

local discoveredOverlays = {}

local overlaysChecked = {}
Cartographer3_Utils.AddEventListener("WORLD_MAP_UPDATE", function(event)
	local currentMapTexture = GetMapInfo()
	if Cartographer3_Data.currentMapTexture == currentMapTexture then
		return
	end
	Cartographer3_Data.currentContinentID = GetCurrentMapContinent()
	if not currentMapTexture then
		if Cartographer3_Data.currentContinentID == -1 then
			currentMapTexture = "Cosmic"
		else
			currentMapTexture = "World"
		end
	end
	Cartographer3_Data.currentMapTexture = currentMapTexture
	Cartographer3_Data.currentZoneData = Cartographer3_Data.ZONE_DATA[currentMapTexture]
	Cartographer3_Data.currentContinentData = Cartographer3_Data.CONTINENT_DATA[Cartographer3_Data.currentContinentID]
	
	if Cartographer3_Data.currentZoneData then
		local scale = Cartographer3_Data.currentZoneData.mapWidth / BLIZZARD_MAP_WIDTH
		_G.WorldMapDetailFrame:SetScale(scale)
		_G.WorldMapButton:SetScale(scale)
		WorldMapFrame:SetScale(scale)
		_G.WorldMapPositioningGuide:SetScale(scale)
		_G.WorldMapDetailFrame:SetPoint("BOTTOMLEFT", mapView, "CENTER", Cartographer3_Data.currentZoneData.mapLeft / scale, Cartographer3_Data.currentZoneData.mapBottom / scale)
	elseif Cartographer3_Data.currentContinentData and Cartographer3_Data.currentContinentData[1] then
		local scale = Cartographer3_Data.currentContinentData.width / BLIZZARD_MAP_WIDTH
		_G.WorldMapDetailFrame:SetScale(scale)
		_G.WorldMapButton:SetScale(scale)
		WorldMapFrame:SetScale(scale)
		_G.WorldMapPositioningGuide:SetScale(scale)
		_G.WorldMapDetailFrame:SetPoint("CENTER", mapView, "CENTER", Cartographer3_Data.currentContinentData.mapCenter / scale, Cartographer3_Data.currentContinentData.mapMiddle / scale)
	end
	
--	if mapView then
--		Cartographer3:PLAYER_DEAD()
--	end
	
	if overlaysChecked[currentMapTexture] then
		return
	end
	overlaysChecked[currentMapTexture] = true
	
	local numOverlays = GetNumMapOverlays()
	local len = #currentMapTexture + 21 -- #([=[Interface\WorldMap\]=] .. currentMapTexture .. [=[\]=])
	for i = 1, numOverlays do
		local tname, tw, th, ofx, ofy = GetMapOverlayInfo(i)
		tname = tname:sub(len)
		if not discoveredOverlays[tname] then
			local num = tw + th * 1024 + ofx * 1048576 + ofy * 1073741824
			if num ~= 0 and num ~= 131200 and tname ~= "" and tname:lower() ~= "pixelfix" then
				discoveredOverlays[tname] = num
				local OVERLAY_DATA__currentMapTexture = Cartographer3_Data.OVERLAY_DATA[currentMapTexture]
				if not OVERLAY_DATA__currentMapTexture then
					OVERLAY_DATA__currentMapTexture = {}
					Cartographer3_Data.OVERLAY_DATA[currentMapTexture] = OVERLAY_DATA__currentMapTexture
				end
				if OVERLAY_DATA__currentMapTexture[tname] ~= num then
					OVERLAY_DATA__currentMapTexture[tname] = num
					local Cartographer3_NewOverlaysDB = _G.Cartographer3_NewOverlaysDB
					if type(Cartographer3_NewOverlaysDB) ~= "table" then
						Cartographer3_NewOverlaysDB = {}
						_G.Cartographer3_NewOverlaysDB = Cartographer3_NewOverlaysDB
					end
					if not Cartographer3_NewOverlaysDB[currentMapTexture] then
						Cartographer3_NewOverlaysDB[currentMapTexture] = {}
					end
					Cartographer3_NewOverlaysDB[currentMapTexture][tname] = num
				end
			end
		end
	end
end)

local MATCH_ERR_ZONE_EXPLORED = "^" .. ERR_ZONE_EXPLORED:gsub("%%s", ".+") .. "$"
Cartographer3_Utils.AddEventListener("UI_INFO_MESSAGE", function(event, message)
	if message:match(MATCH_ERR_ZONE_EXPLORED) then
		local fileName = Cartographer3_Data.LOCALIZED_ZONE_TO_TEXTURE[GetRealZoneText()]
		if fileName then
			overlaysChecked[fileName] = nil
		end
	end
end)

do
	local function f()
		Cartographer3_Utils.RemoveTimer(f)
		if db.profile.followPlayer then
			SetMapToCurrentZone()
			Cartographer3_Utils.ZoomToPlayer()
		else
			Cartographer3.Utils.ZoomToCurrentZone()
		end
	end
	Cartographer3_Utils.AddEventListener("ZONE_CHANGED_NEW_AREA", function(event)
		Cartographer3_Utils.AddTimer(f)
	end)
end

local instanceTextureFrames = {}
Cartographer3_Data.instanceTextureFrames = instanceTextureFrames
Cartographer3_Data.instanceTop = 0
Cartographer3_Data.instanceLeft = 0
Cartographer3_Data.instanceBottom = 0
Cartographer3_Data.instanceRight = 0
Cartographer3_Data.instanceRows = 0

local zoneSearchData = {} -- data in here is split up by continent and then into a 10x10 grid of possible zones, in the order they should be searched (by city and then by area)
Cartographer3_Data.zoneSearchData = zoneSearchData

Cartographer3_Data.cameraX = 0
Cartographer3_Data.cameraY = 0
Cartographer3_Data.cameraZoom = 1

local function initializeContinentData()
	initializeContinentData = nil
	local continentNames = { GetMapContinents() }
	for i,v in ipairs(continentNames) do
		local data = Cartographer3_Data.CONTINENT_DATA[i]
		if not data then
			-- expansion we don't know about
			data = {
				x = 0,
				y = 375,
				width = 225,
				rect = { 0, 1, 0, 1 }
			}
			Cartographer3_Data.CONTINENT_DATA[i] = data
		end
		data.name = v
	end
	for continentID, data in ipairs(Cartographer3_Data.CONTINENT_DATA) do
		local currentContinentData = {}
		Cartographer3_Data.CONTINENT_DATA[continentID] = currentContinentData
		SetMapZoom(continentID)
		currentContinentData.texture = Cartographer3_Data.currentMapTexture
		Cartographer3_Data.CONTINENT_DATA[Cartographer3_Data.currentMapTexture] = currentContinentData
		currentContinentData.id = continentID
		
		currentContinentData.localizedName = data.name
		
		local continentCenter = data.x
		currentContinentData.mapCenter = continentCenter
		
		local continentMiddle = data.y
		currentContinentData.mapMiddle = continentMiddle
		
		local continentWidth = data.yards / Cartographer3_Data.YARDS_PER_PIXEL
		currentContinentData.width = continentWidth
		
		local continentHeight = continentWidth * BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO
		currentContinentData.height = continentHeight
		
		local rect = data.rect
		currentContinentData.visibleLeft = continentCenter + continentWidth * (rect[1] - 0.5)
		currentContinentData.visibleRight = continentCenter + continentWidth * (rect[2] - 0.5)
		currentContinentData.visibleTop = continentMiddle - continentHeight * (rect[3] - 0.5)
		currentContinentData.visibleBottom = continentMiddle - continentHeight * (rect[4] - 0.5)
		
		local zoneNames = { GetMapZones(continentID) }
		local zones = {}
		currentContinentData.zones = zones
		for _ = 1, #zoneNames do
			local x, y
			local name, fileName, texPctX, texPctY, texX, texY, scrollX, scrollY
			
			repeat
				x, y = math.random(), math.random()
				name, fileName, texPctX, texPctY, texX, texY, scrollX, scrollY = UpdateMapHighlight(x, y)
			until fileName and not zones[fileName]
			
			local zones_fileName = {}
			zones_fileName.continentID = continentID
			for i, n in ipairs(zoneNames) do
				if n == name then
					zones_fileName.id = i
				end
			end
			zones_fileName.texture = fileName
			zones_fileName.localizedName = name
			zones_fileName.mapWidth = texX
			zones_fileName.mapHeight = texY
			zones_fileName.mapLeft = scrollX
			zones_fileName.mapBottom = scrollY
			Cartographer3_Data.ZONE_DATA[fileName] = zones_fileName
			zones[fileName] = zones_fileName
			Cartographer3_Data.LOCALIZED_ZONE_TO_TEXTURE[name] = fileName
			
			for x = scrollX, scrollX+texX, 0.0025 do
				for y = scrollY, scrollY+texY, 0.0025 do
					if UpdateMapHighlight(x, y) == name then
						if not zones_fileName.highlightLeft then
							zones_fileName.highlightLeft = x
							zones_fileName.highlightRight = x
							zones_fileName.highlightBottom = y
							zones_fileName.highlightTop = y
						else
							if x < zones_fileName.highlightLeft then
								zones_fileName.highlightLeft = x
							elseif x > zones_fileName.highlightRight then
								zones_fileName.highlightRight = x
							end
							if y < zones_fileName.highlightBottom then
								zones_fileName.highlightBottom = y
							elseif y > zones_fileName.highlightTop then
								zones_fileName.highlightTop = y
							end
						end
					end
				end
			end
			
			zones_fileName.highlightLeft = zones_fileName.highlightLeft-0.001
			zones_fileName.highlightRight = zones_fileName.highlightRight+0.001
			zones_fileName.highlightBottom = zones_fileName.highlightBottom-0.001
			zones_fileName.highlightTop = zones_fileName.highlightTop+0.001
			zones_fileName.visibleLeft = zones_fileName.mapLeft
			zones_fileName.visibleBottom = zones_fileName.mapBottom
			if fileName == "EversongWoods" or fileName == "Ghostlands" or fileName == "Sunwell" or fileName == "SilvermoonCity" then
				zones_fileName.mapLeft = zones_fileName.mapLeft - 0.00168
				zones_fileName.mapBottom = zones_fileName.mapBottom + 0.01
				zones_fileName.highlightLeft = zones_fileName.highlightLeft - 0.00168
				zones_fileName.highlightRight = zones_fileName.highlightRight - 0.00168
				zones_fileName.highlightBottom = zones_fileName.highlightBottom + 0.01
				zones_fileName.highlightTop = zones_fileName.highlightTop + 0.01
			end
			
			zones_fileName.mapBottom = 1 - zones_fileName.mapBottom - zones_fileName.mapHeight
			zones_fileName.visibleBottom = 1 - zones_fileName.visibleBottom - zones_fileName.mapHeight
			zones_fileName.mapWidth = zones_fileName.mapWidth * continentWidth
			zones_fileName.mapHeight = zones_fileName.mapHeight * continentHeight
			zones_fileName.mapLeft = zones_fileName.mapLeft * continentWidth - continentWidth/2 + continentCenter
			zones_fileName.mapBottom = zones_fileName.mapBottom * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName.visibleLeft = zones_fileName.visibleLeft * continentWidth - continentWidth/2 + continentCenter
			zones_fileName.visibleBottom = zones_fileName.visibleBottom * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName.highlightLeft = zones_fileName.highlightLeft * continentWidth - continentWidth/2 + continentCenter
			zones_fileName.highlightRight = zones_fileName.highlightRight * continentWidth - continentWidth/2 + continentCenter
			zones_fileName.highlightBottom, zones_fileName.highlightTop = 1 - zones_fileName.highlightTop, 1 - zones_fileName.highlightBottom
			zones_fileName.highlightBottom = zones_fileName.highlightBottom * continentHeight - continentHeight/2 + continentMiddle
			zones_fileName.highlightTop = zones_fileName.highlightTop * continentHeight - continentHeight/2 + continentMiddle
		end
		
		currentContinentData.minimapTextureCenterX = data.minimapTextureCenterX
		currentContinentData.minimapTextureCenterY = data.minimapTextureCenterY
		currentContinentData.minimapTextureOffsetX = data.minimapTextureOffsetX
		currentContinentData.minimapTextureOffsetY = data.minimapTextureOffsetY
		currentContinentData.minimapTextureBackground = data.minimapTextureBackground
		local extraMinimapTextureAreas = {}
		currentContinentData.extraMinimapTextureAreas = extraMinimapTextureAreas
		if data.extraMinimapTextureAreas then
			for k,v in pairs(data.extraMinimapTextureAreas) do
				extraMinimapTextureAreas[#extraMinimapTextureAreas+1] = {
					k,
					v.minimapTextureCenterX,
					v.minimapTextureCenterY,
				}
			end
		end
		
		currentContinentData.rectLeft = rect[1]
		currentContinentData.rectRight = rect[2]
		currentContinentData.rectTop = rect[3]
		currentContinentData.rectBottom = rect[4]
	end
	SetMapToCurrentZone()
	
	local function zoneSearchData_func(alpha, bravo)
		if Cartographer3_Data.CITIES[bravo] then
			return false
		elseif Cartographer3_Data.CITIES[alpha] then
			return true
		end
		
		local zoneData_alpha = Cartographer3_Data.ZONE_DATA[alpha]
		local zoneData_bravo = Cartographer3_Data.ZONE_DATA[bravo]
		local area_alpha = (zoneData_alpha.highlightRight - zoneData_alpha.highlightLeft) * (zoneData_alpha.highlightTop - zoneData_alpha.highlightBottom)
		local area_bravo = (zoneData_bravo.highlightRight - zoneData_bravo.highlightLeft) * (zoneData_bravo.highlightTop - zoneData_bravo.highlightBottom)
		return area_alpha < area_bravo
	end
	for continentID, currentContinentData in ipairs(Cartographer3_Data.CONTINENT_DATA) do
		local t = {}
		zoneSearchData[continentID] = t
		local zones = currentContinentData.zones
		
		local continentLeft = currentContinentData.visibleLeft
		local continentWidth = currentContinentData.visibleRight - currentContinentData.visibleLeft
		local continentBottom = currentContinentData.visibleBottom
		local continentHeight = currentContinentData.visibleTop - currentContinentData.visibleBottom
		
		for fileName, zoneData in pairs(zones) do
			local zoneLeft = (zoneData.highlightLeft - continentLeft)/continentWidth
			local zoneRight = (zoneData.highlightRight - continentLeft)/continentWidth
			local zoneBottom = (zoneData.highlightBottom - continentBottom)/continentHeight
			local zoneTop = (zoneData.highlightTop - continentBottom)/continentHeight
			for x = 0, 9 do
				if zoneLeft < (x+1)/10 and zoneRight > x/10 then
					for y = 0, 9 do
						if zoneBottom < (y+1)/10 and zoneTop > y/10 then
							local num = x*10 + y
							local u = t[num]
							if not u then
								u = {}
								t[num] = u
							end
							u[#u+1] = fileName
						end
					end
				end
			end
		end
		for num = 0, 99 do
			if t[num] then
				table.sort(t[num], zoneSearchData_func)
			end
		end
	end
end

local mapTextures = {}
local function createMapTextures()
	createMapTextures = nil
	for continentID, currentContinentData in ipairs(Cartographer3_Data.CONTINENT_DATA) do
		local continentCenter = currentContinentData.mapCenter
		local continentMiddle = currentContinentData.mapMiddle
		local continentWidth = currentContinentData.width
		local continentHeight = currentContinentData.height
		local continentTileSize = continentWidth * 256/BLIZZARD_MAP_WIDTH
		local rectLeft = currentContinentData.rectLeft
		local rectRight = currentContinentData.rectRight
		local rectTop = currentContinentData.rectTop
		local rectBottom = currentContinentData.rectBottom
		for i = 1, 12 do
			local good = true
			local left = currentContinentData.rectLeft*1002/1024 * 4 - ((i-1)%4)
			if left < 0 then
				left = 0
			elseif left >= 1 then
				good = false
			end
			local right = currentContinentData.rectRight*1002/1024 * 4 - ((i-1)%4)
			if right > 1 then
				right = 1
			elseif right <= 0 then
				good = false
			end
			local top = currentContinentData.rectTop*668/768 * 3 - math.floor((i-1) / 4)
			if top < 0 then
				top = 0
			elseif top >= 1 then
				good = false
			end
			local bottom =  currentContinentData.rectBottom*668/768 * 3 - math.floor((i-1) / 4)
			if bottom > 1 then
				bottom = 1
			elseif bottom <= 0 then
				good = false
			end
			
			if good then
				local x = ((i-1) % 4) * continentTileSize + continentCenter - continentWidth/2
				local y = -math.floor((i-1) / 4 + 1) * continentTileSize + continentMiddle + continentHeight/2
				local x1 = x + left * continentTileSize
				local x2 = x + right * continentTileSize
				local y1 = y + (1 - bottom) * continentTileSize
				local y2 = y + (1 - top) * continentTileSize
				
				local tex = mapView:CreateTexture(nil--[[mapView:GetName() .. "_MapTexture_" .. currentContinentData.texture .. i]], "BORDER")
				mapTextures[#mapTextures+1] = tex
				local mapFileName = currentContinentData.texture
				tex:SetTexture(([=[Interface\WorldMap\%s\%s%d]=]):format(mapFileName, mapFileName, i))
				tex:SetPoint("BOTTOMLEFT", mapView, "CENTER", x1, y1)
				tex:SetPoint("TOPRIGHT", mapView, "CENTER", x2, y2)
			
				tex:SetTexCoord(left, right, top, bottom)
			end
		end
		
		-- no longer use rect data
		currentContinentData.rectLeft = nil
		currentContinentData.rectRight = nil
		currentContinentData.rectTop = nil
		currentContinentData.rectBottom = nil
	end
end

local cos, sin, abs = cos, sin, math.abs

local totalPOIs = {}
Cartographer3_Data.totalPOIs = totalPOIs
local shownPOIs = {}
local function poi_OnShow(this)
	shownPOIs[this] = true
	
	local size = 20 / Cartographer3_Data.cameraZoom
	local total_size = size * totalPOIs[this]
	this:SetWidth(total_size)
	this:SetHeight(total_size)
end
local function poi_OnHide(this)
	shownPOIs[this] = nil
end
local inPOI = nil
local function poi_OnEnter(this)
	inPOI = 0
end
local function poi_OnLeave(this)
	inPOI = 0
end

local tmp = {}
local function poi_sort(alpha, bravo)
	return alpha:GetName() < bravo:GetName()
end
local alreadySetOwner = false
function Cartographer3.UpdatePOITooltips(elapsed, currentTime)
	if not inPOI or inPOI > currentTime then
		return
	end
	inPOI = currentTime + 0.5
	if not alreadySetOwner then
		WorldMapTooltip:SetOwner(mapHolder, "ANCHOR_CURSOR")
		alreadySetOwner = true
	end
	for poi in pairs(shownPOIs) do
		if Cartographer3_Utils.IsMouseHovering(poi) then
			tmp[#tmp+1] = poi
		end
	end
	WorldMapTooltip:ClearLines()
	if #tmp == 0 then
		inPOI = nil
		alreadySetOwner = false
		WorldMapTooltip:Hide()
	elseif #tmp == 1 then
		tmp[1]:AddDataToFullTooltip()
		tmp[1] = nil
	else
		table.sort(tmp, poi_sort)
		for i, v in ipairs(tmp) do
			v:AddDataToTooltipLine()
			tmp[i] = nil
		end
	end
	WorldMapTooltip:Show()
end
setmetatable(totalPOIs, {__newindex=function(self, poi, scale)
	rawset(self, poi, scale)
	if poi:IsShown() then
		shownPOIs[poi] = true
	end
	poi:SetScript("OnShow", poi_OnShow)
	poi:SetScript("OnHide", poi_OnHide)
	poi:SetScript("OnEnter", poi_OnEnter)
	poi:SetScript("OnLeave", poi_OnLeave)
	poi:EnableMouse(true)
	
	local size = 20 / Cartographer3_Data.cameraZoom
	local total_size = size * scale
	poi:SetWidth(total_size)
	poi:SetHeight(total_size)
end})

function Cartographer3.AddPOI(poi, scale)
	totalPOIs[poi] = scale or 1
end

local loadInstanceTextureSection
do
	local doneInstances = {}
	local workingInstance, workingInstancePosition
	local first = true
	
	local function layout()
		local names = {}
		local widest = 0
		for name, instanceTextureFrame in pairs(instanceTextureFrames) do
			local angle = Cartographer3_Data.INSTANCE_ROTATIONS[name] or 0
			angle = angle - 90
			local ULx, ULy, LLx, LLy, URx, URy, LRx, LRy = Cartographer3_Utils.RotateTexCoord(angle)
			
			local A, B = abs(cos(angle)), abs(sin(angle))
			
			local left, right, bottom, top
			for i, tex in ipairs(instanceTextureFrame) do
				tex.x, tex.y = Cartographer3_Utils.RotateCoordinate(angle, tex.x + tex.w/2, tex.y + tex.h/2)
				tex.w, tex.h = tex.w*A + tex.h*B, tex.w*B + tex.h*A
				tex:SetWidth(tex.w)
				tex:SetHeight(tex.h)
				tex:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy)
				local tex_left = tex.x - tex.w/2
				local tex_right = tex.x + tex.w/2
				local tex_bottom = tex.y - tex.h/2
				local tex_top = tex.y + tex.h/2
				if not left then
					left, right, bottom, top = tex_left, tex_right, tex_bottom, tex_top
				else
					if tex_left < left then
						left = tex_left
					elseif tex_right > right then
						right = tex_right
					end
					if tex_bottom < bottom then
						bottom = tex_bottom
					elseif tex_top > top then
						top = tex_top
					end
				end
			end
			local width = right - left
			local height = top - bottom
			local center = left + width/2
			local middle = bottom + height/2
			for i, tex in ipairs(instanceTextureFrame) do
				tex.x = tex.x - center
				tex.y = tex.y - middle
				tex:SetPoint("CENTER", instanceTextureFrame, "CENTER", tex.x, tex.y)
			end
			local full_width = width
			local full_height = height
			if width*2/3 < height then
				full_width = height * 3/2
			else
				full_height = width * 2/3
			end
			if full_width > widest then
				widest = full_width
			end
			instanceTextureFrame:SetWidth(full_width)
			instanceTextureFrame:SetHeight(full_height)
			instanceTextureFrame.w = full_width
			instanceTextureFrame.h = full_height
			names[#names+1] = name
		end
		Cartographer3_Data.INSTANCE_ROTATIONS = nil
		table.sort(names)
		local rows = math.ceil((#names)^0.5)
		local separation = 0.5
		separation = separation + 1
		local instances_center = 45000
		local instances_middle = 0
		for i, name in ipairs(names) do
			local pos_x = (i-1) % rows
			local pos_y = math.floor((i-1) / rows)
			local instanceTextureFrame = instanceTextureFrames[name]
			local x = instances_center + (pos_x - rows/2)*widest*separation
			local y = instances_middle - (pos_y - rows/2)*widest*2/3*separation
			instanceTextureFrame.x = x
			instanceTextureFrame.y = y
			instanceTextureFrame:SetPoint("CENTER", mapView, "CENTER", x, y)
			instanceTextureFrames[#instanceTextureFrames+1] = instanceTextureFrame
			
			local f = CreateFrame("Frame", nil, mapView)
			f:SetScale(instanceTextureFrame:GetScale())
			f:SetWidth(instanceTextureFrame:GetWidth())
			f:SetHeight(instanceTextureFrame:GetHeight())
			f:SetPoint("CENTER", mapView, "CENTER", x, y)
			local background = f:CreateTexture(nil--[[instanceTextureFrame:GetName() .. "_Background"]], "BACKGROUND")
			background:SetTexture(0, 0, 0)
			background:SetAllPoints(f)
			instanceTextureFrame:SetParent(nil)
			instanceTextureFrame:SetParent(mapView)
		end
		Cartographer3_Data.instanceLeft = (instances_center - (rows + 1)/2*widest*separation) * 0.01
		Cartographer3_Data.instanceRight = (instances_center + (rows - 1)/2*widest*separation) * 0.01
		Cartographer3_Data.instanceBottom = (instances_middle - (rows - 1)/2*widest*2/3*separation) * 0.01
		Cartographer3_Data.instanceTop = (instances_middle + (rows + 1)/2*widest*2/3*separation) * 0.01
		Cartographer3_Data.instanceRows = rows
		Cartographer3_Utils.RemoveTimer(layout)
		layout = nil
		Cartographer3_Utils.CollectGarbageSoon()
	end
	local function getSizes()
		local hasBad = false
		for name, instanceTextureFrame in pairs(instanceTextureFrames) do
			local good = true
			for i, tex in ipairs(instanceTextureFrame) do
				if not tex.w then
					local w = tex:GetWidth()
					if w == 0 then
						good = false
						break
					end
					tex.w = w
					tex.h = tex:GetHeight()
				end
			end
			if not good then
				hasBad = name
			end
		end
		if not hasBad then
			Cartographer3_Utils.RemoveTimer(getSizes)
			getSizes = nil
			
			Cartographer3_Utils.AddTimer(layout, true)
		end
	end
	function loadInstanceTextureSection()
		if first then
			first = false
			for name, data in pairs(Cartographer3_Data.INSTANCE_TEXTURE_DATA) do
				instanceTextureFrame = CreateFrame("Frame", mapView:GetName() .. "_InstanceTextureFrame_" .. name:gsub("[ ']", ""), mapView)
				instanceTextureFrames[name] = instanceTextureFrame
				instanceTextureFrame.name = name
				local size = 0.01
				instanceTextureFrame:SetScale(0.01)
				local offsetX = (200 + math.random()*400) / size-- -currentContinentData.minimapTextureOffsetX + currentContinentData.mapCenter
				local offsetY = (-200 + math.random()*400) / size-- currentContinentData.minimapTextureOffsetY + currentContinentData.mapMiddle
				instanceTextureFrame:SetPoint("CENTER", mapView, "CENTER", offsetX, offsetY)
				instanceTextureFrame:SetWidth(1e-10)
				instanceTextureFrame:SetHeight(1e-10)
				instanceTextureFrame:Hide()
				instanceTextureFrame:SetAlpha(0)
			end
			return
		end	
		if not workingInstance then
			for name, data in pairs(Cartographer3_Data.INSTANCE_TEXTURE_DATA) do
				if not doneInstances[name] then
					workingInstance = name
				end
			end
			if not workingInstance then
				-- all done
				Cartographer3_Utils.RemoveTimer(loadInstanceTextureSection)
				loadInstanceTextureSection = nil
				Cartographer3_Data.INSTANCE_TEXTURE_DATA = nil
				Cartographer3_Utils.AddTimer(getSizes, true)
				return
			end
			workingInstancePosition = 0
		end
		
		local workingInstanceData = Cartographer3_Data.INSTANCE_TEXTURE_DATA[workingInstance]
		local instanceTextureFrame = instanceTextureFrames[workingInstance]
		
		local finish = GetTime() + Cartographer3_Data.ITERATIVE_PROCESS_STEP_SECONDS
		while GetTime() < finish do
			local offset = workingInstancePosition * 28
			if offset >= #workingInstanceData then
				workingInstancePosition = nil
				break
			end
			workingInstancePosition = workingInstancePosition + 1
			
			local texturePath = ("%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"):format(workingInstanceData:byte(offset + 1, offset + 16))
			
			local x = Cartographer3_Utils.UnpackFloat(workingInstanceData:byte(offset + 17, offset + 20))
			local y = Cartographer3_Utils.UnpackFloat(workingInstanceData:byte(offset + 21, offset + 24))
			local z = Cartographer3_Utils.UnpackFloat(workingInstanceData:byte(offset + 25, offset + 28))
			
			local tex = instanceTextureFrame:CreateTexture(nil--[[minimapStyleTextureFrame:GetName() .. "_Texture" .. x .. "_" .. y .. "_" .. z]], "ARTWORK")
			instanceTextureFrame[#instanceTextureFrame+1] = tex
			tex:SetTexture(([=[textures\Minimap\%s]=]):format(texturePath))
			tex.x = x
			tex.y = y
			tex.z = z
			tex:SetWidth(0)
			tex:SetHeight(0)
		end
		if not workingInstancePosition then
			Cartographer3_Data.INSTANCE_TEXTURE_DATA[workingInstance] = nil
			doneInstances[workingInstance] = true
			workingInstance = nil
		end
	end
end

local minimapStyleTextureFrames = {}
local loadMinimapStyleTextureSection
do
	local doneContinents = {}
	local workingContinent, workingExtraArea, workingTextureDataPosition
	local first = true
	function loadMinimapStyleTextureSection()
		if first then
			first = false
			for i, currentContinentData in ipairs(Cartographer3_Data.CONTINENT_DATA) do
				local currentContinentData_texture = currentContinentData.texture
				minimapStyleTextureFrame = CreateFrame("Frame", mapView:GetName() .. "_MinimapStyleTextureFrame_" .. currentContinentData_texture, mapView)
				minimapStyleTextureFrames[currentContinentData_texture] = minimapStyleTextureFrame
				local offsetX = -currentContinentData.minimapTextureOffsetX + currentContinentData.mapCenter
				local offsetY = currentContinentData.minimapTextureOffsetY + currentContinentData.mapMiddle
				minimapStyleTextureFrame:SetPoint("CENTER", mapView, "CENTER", offsetX, offsetY)
				minimapStyleTextureFrame:SetWidth(1e-10)
				minimapStyleTextureFrame:SetHeight(1e-10)
				if currentContinentData.minimapTextureBackground then
					local tex = minimapStyleTextureFrame:CreateTexture(nil--[[minimapStyleTextureFrame:GetName() .. "_Background"]], "BACKGROUND")
					tex:SetTexture(([=[textures\Minimap\%s]=]):format(currentContinentData.minimapTextureBackground))
					tex:SetPoint("BOTTOMLEFT", minimapStyleTextureFrame, "CENTER", currentContinentData.visibleLeft - offsetX, currentContinentData.visibleBottom - offsetY)
					tex:SetPoint("TOPRIGHT", minimapStyleTextureFrame, "CENTER", currentContinentData.visibleRight - offsetX, currentContinentData.visibleTop - offsetY)
				end
				minimapStyleTextureFrame:Hide()
				
			end
			return
		end
		if not workingContinent then
			if Cartographer3_Data.CONTINENT_DATA[Cartographer3_Data.currentMapTexture] then
				workingContinent = Cartographer3_Data.CONTINENT_DATA[Cartographer3_Data.currentMapTexture].id
			elseif Cartographer3_Data.currentZoneData then
				workingContinent = Cartographer3_Data.currentZoneData.continentID
			end
			
			local isDone
			if workingContinent then
				local continentTexture = Cartographer3_Data.CONTINENT_DATA[workingContinent].texture
				isDone = doneContinents[continentTexture]
			end	
			if isDone or not workingContinent then
				workingContinent = nil
				for continentID, currentContinentData in ipairs(Cartographer3_Data.CONTINENT_DATA) do
					if not doneContinents[currentContinentData.texture] then
						workingContinent = continentID
					end
				end
			end
			if not workingContinent then
				-- all done
				Cartographer3_Utils.RemoveTimer(loadMinimapStyleTextureSection)
				Cartographer3_Data.TEXTURE_DATA = nil
				loadMinimapStyleTextureSection = nil
				return
			end
		end
	
		local currentContinentData = Cartographer3_Data.CONTINENT_DATA[workingContinent]
		local currentContinentData_texture = currentContinentData.texture
		local minimapStyleTextureFrame = minimapStyleTextureFrames[currentContinentData_texture]
		
		local size = BLIZZARD_MINIMAP_TILE_YARD_SIZE / Cartographer3_Data.YARDS_PER_PIXEL
		local centerX, centerY
		
		local currentTextureData
		if not workingExtraArea then
			currentTextureData = Cartographer3_Data.TEXTURE_DATA[currentContinentData_texture]
			centerX, centerY = currentContinentData.minimapTextureCenterX, currentContinentData.minimapTextureCenterY
		else
			local extraMinimapTextureArea = currentContinentData.extraMinimapTextureAreas[workingExtraArea]
			currentTextureData = Cartographer3_Data.TEXTURE_DATA[extraMinimapTextureArea[1]]
			centerX, centerY = extraMinimapTextureArea[2], extraMinimapTextureArea[3]
		end
	
		local finish = GetTime() + Cartographer3_Data.ITERATIVE_PROCESS_STEP_SECONDS
		while GetTime() < finish do
			local texturePath
			workingTextureDataPosition, texturePath = next(currentTextureData, workingTextureDataPosition)
			if not workingTextureDataPosition then
				break
			end
			local x, y = math.floor(workingTextureDataPosition / 100), workingTextureDataPosition % 100
			local chunk_num = math.floor(x - centerX) * 100 + math.floor(y - centerY) + 10000
			local tex = minimapStyleTextureFrame:CreateTexture(nil--[[minimapStyleTextureFrame:GetName() .. "_Texture" .. chunk_num]], "ARTWORK")
			minimapStyleTextureFrame[chunk_num] = tex
			tex:Hide()
			tex:SetTexture(([=[textures\Minimap\%s]=]):format(texturePath))
			tex:SetWidth(size)
			tex:SetHeight(size)
			tex:SetAlpha(1)
			tex:SetPoint("BOTTOMLEFT", minimapStyleTextureFrame, "CENTER", (x - centerX) * size, -(y - centerY) * size)
		end
		if not workingTextureDataPosition then
			Cartographer3_Data.TEXTURE_DATA[currentContinentData.texture] = nil
			doneContinents[currentContinentData.texture] = true
			if not workingExtraArea then
				workingExtraArea = 1
			else
				workingExtraArea = workingExtraArea + 1
			end
			
			if not currentContinentData.extraMinimapTextureAreas[workingExtraArea] then
				-- done continent
				workingContinent = nil
				workingExtraArea = nil
				-- no longer using minimap texture data, clear it, refresh the table
				currentContinentData.minimapTextureCenterX = nil
				currentContinentData.minimapTextureCenterY = nil
				currentContinentData.minimapTextureOffsetX = nil
				currentContinentData.minimapTextureOffsetY = nil
				currentContinentData.minimapTextureBackground = nil
				currentContinentData.extraMinimapTextureAreas = nil
			end
		end
	end
end

local zoneFrames = {}
local loadMapOverlayTextureSection
do
	local workingZone, workingTName
	local doneZones = {}
	local first = true
	function loadMapOverlayTextureSection()
		if first then
			first = false
			
			local order = {}
			for zoneTexture in pairs(Cartographer3_Data.ZONE_DATA) do
				order[#order+1] = zoneTexture
			end
			table.sort(order, function(alpha, bravo)
				if Cartographer3_Data.CITIES[alpha] then
					return false
				elseif Cartographer3_Data.CITIES[bravo] then
					return true
				else
					return alpha > bravo
				end
			end)
			
			for _, zoneTexture in ipairs(order) do
				local zoneData = Cartographer3_Data.ZONE_DATA[zoneTexture]
				local zoneFrame = CreateFrame("Frame", mapView:GetName() .. "_ZoneFrame_" .. zoneTexture, mapView)
				zoneFrame.texture = zoneTexture
				zoneFrame:Hide()
				zoneFrame:SetAlpha(0)
				zoneFrame:SetWidth(BLIZZARD_MAP_WIDTH)
				zoneFrame:SetHeight(BLIZZARD_MAP_WIDTH * BLIZZARD_MAP_HEIGHT_TO_WIDTH_RATIO)
				zoneFrames[zoneTexture] = zoneFrame
				local scale = zoneData.mapWidth / BLIZZARD_MAP_WIDTH
				zoneFrame:SetScale(scale)
				zoneFrame:SetPoint("BOTTOMLEFT", mapView, "CENTER", zoneData.mapLeft / scale, zoneData.mapBottom / scale)
			end
			return
		end
		if not workingZone then
			if Cartographer3_Data.currentZoneData then
				workingZone = Cartographer3_Data.currentMapTexture
			end
			
			if not workingZone or doneZones[workingZone] then
				workingZone = nil
				for zoneTexture in pairs(Cartographer3_Data.ZONE_DATA) do
					if not doneZones[zoneTexture] then
						workingZone = zoneTexture
						break
					end
				end
				if not workingZone then
					-- all done
					Cartographer3_Utils.RemoveTimer(loadMapOverlayTextureSection)
					loadMapOverlayTextureSection = nil
					return
				end
			end
		end
		local zoneFrame = zoneFrames[workingZone]
		assert(zoneFrame)
		local overlayData = Cartographer3_Data.OVERLAY_DATA[workingZone]
		doneZones[workingZone] = true
		
		if overlayData then
			local finishTime = GetTime() + Cartographer3_Data.ITERATIVE_PROCESS_STEP_SECONDS
			while GetTime() < finishTime do
				local num
				workingTName, num = next(overlayData, workingTName)
				if not workingTName then
					break
				end
				
				local textureName = ([=[Interface\WorldMap\%s\%s]=]):format(workingZone, workingTName)
				local textureWidth = num % 1024
				local textureHeight = math.floor(num / 1024) % 1024
				local offsetX = math.floor(num / (1024*1024)) % 1024
				local offsetY = math.floor(num / (1024*1024*1024)) % 1024

				if textureName == [=[Interface\WorldMap\Tirisfal\BRIGHTWATERLAKE]=] then
					if offsetX == 587 then
						offsetX = 584
					end
				elseif textureName == [=[Interface\WorldMap\Silverpine\BERENSPERIL]=] then
					if offsetY == 417 then
						offsetY = 415
					end
				end

				local numTexturesWide = math.ceil(textureWidth / 256)
				local numTexturesTall = math.ceil(textureHeight / 256)

				for j = 1, numTexturesTall do
					local texturePixelHeight
					local textureFileHeight
					if j < numTexturesTall then
						texturePixelHeight = 256
						textureFileHeight = 256
					else
						texturePixelHeight = textureHeight % 256
						if texturePixelHeight == 0 then
							texturePixelHeight = 256
						end
						textureFileHeight = 16
						while textureFileHeight < texturePixelHeight do
							textureFileHeight = textureFileHeight * 2
						end
					end
					for k = 1, numTexturesWide do
						local texturePixelWidth
						local textureFileWidth
						if k < numTexturesWide then
							texturePixelWidth = 256
							textureFileWidth = 256
						else
							texturePixelWidth = textureWidth % 256
							if texturePixelWidth == 0 then
								texturePixelWidth = 256
							end
							textureFileWidth = 16
							while textureFileWidth < texturePixelWidth do
								textureFileWidth = textureFileWidth * 2
							end
						end

						local texture = zoneFrame:CreateTexture(nil--[[zoneFrame:GetName() .. "_Texture" .. (#zoneFrame+1)]], "ARTWORK")
						zoneFrame[#zoneFrame+1] = texture
						texture:SetWidth(texturePixelWidth)
						texture:SetHeight(texturePixelHeight)
						texture:SetTexCoord(0, texturePixelWidth/textureFileWidth, 0, texturePixelHeight/textureFileHeight)
						texture:SetPoint("TOPLEFT", zoneFrame, "TOPLEFT", (offsetX + (256 * (k-1))), -(offsetY + (256 * (j - 1))))
						texture:SetTexture(textureName..(((j - 1) * numTexturesWide) + k))

						-- if discovered[tname] then
						-- 	texture:SetVertexColor(1.0,1.0,1.0)
						-- 	texture:SetAlpha(1.0)
						-- else
						-- 	texture:SetVertexColor(self.db.profile.darkR, self.db.profile.darkG, self.db.profile.darkB)
						-- 	texture:SetAlpha(self.db.profile.darkA)
						-- end
						texture:Show()
					end
				end
			end
			if not workingTName then
				workingZone = nil
			end
		elseif Cartographer3_Data.CITIES[workingZone] then
			for i = 1, 12 do
				local texture = zoneFrame:CreateTexture(nil--[[zoneFrame:GetName() .. "_Texture" .. (#zoneFrame+1)]], "ARTWORK")
				zoneFrame[#zoneFrame+1] = texture
				texture:SetWidth(256)
				texture:SetHeight(256)
				texture:SetTexture(([=[Interface\WorldMap\%s\%s%d]=]):format(workingZone, workingZone, i))
				texture:SetPoint("TOPLEFT", zoneFrame, "TOPLEFT", ((i - 1) % 4) * 256, -math.floor((i - 1) / 4) * 256)
			end
			workingZone = nil
		else
			workingZone = nil
		end
	end
end

local positionFontString
local timeToUpdate = 0
function Cartographer3.UpdatePositionFontString(elapsed, currentTime)
	timeToUpdate = timeToUpdate - elapsed
	if timeToUpdate > 0 then
		return
	else
		timeToUpdate = 0.1
	end
	local cz, cx, cy
	if Cartographer3_Utils.IsMouseHovering(scrollFrame) then
		cz, cx, cy = Cartographer3_Utils.GetCursorZonePosition()
		if not cx then
			cx, cy = 0, 0
		end
	else
		cz, cx, cy = Cartographer3_Data.currentMapTexture, GetPlayerMapPosition("player")
	end
	
	local zoneData = Cartographer3_Data.ZONE_DATA[cz]
	local zoneName
	if zoneData then
		zoneName = zoneData.localizedName
	else
		local continentData = Cartographer3_Data.CONTINENT_DATA[cz]
		if continentData then
			zoneName = continentData.localizedName
		else
			local instanceTextureFrame = instanceTextureFrames[cz]
			if instanceTextureFrame then
				zoneName = cz
			end
		end
	end
	local subzone
	if Cartographer3_Data.currentMapTexture == cz and (cx ~= 0 or cy ~= 0) then
		subzone = UpdateMapHighlight(cx, cy)
	end
	local formatString
	if zoneName then
		if cx ~= 0 or cy ~= 0 then
			if subzone then
				formatString = "%1$s: %2$s (%3$.1f%5$s%4$.1f)"
			else
				formatString = "%1$s (%3$.1f%5$s%4$.1f)"
			end
		else
			if subzone then
				formatString = "%1$s: %2$s"
			else
				formatString = "%1$s"
			end
		end
	else
		formatString = ""
	end
	if formatString == "" then
		positionFontString:SetText("")
	else
		positionFontString:SetFormattedText(formatString, zoneName, subzone, cx*100, cy*100, Cartographer3_Data.COORDINATE_SEPARATOR)
	end
end

local moving, mouseDown = false, false
function Cartographer3.ShowUIOnHover(elapsed, currentTime)
	local alpha = mapFrame:GetAlpha()
	if moving or mouseDown or Cartographer3_Utils.IsMouseHovering(mapFrame) then
		if alpha == 1 then
			return
		end
		alpha = alpha + elapsed/Cartographer3_Data.FADE_TIME
		if alpha > 1 then
			alpha = 1
		end
	else
		if alpha == 0 then
			return
		end
		alpha = alpha - elapsed/Cartographer3_Data.FADE_TIME
		if alpha < 0 then
			alpha = 0
		end
	end
	mapFrame:SetAlpha(alpha)
end

function Cartographer3.FixViewIfMoving(elapsed, currentTime)
	if moving then
		Cartographer3.Utils.ReadjustCamera()
	end
end

local sizingPoint
function Cartographer3.ChangeCursorIfOnSide(elapsed, currentTime)
	if not mouseDown and mapHolder == GetMouseFocus() then
		local cx, cy = Cartographer3_Utils.GetScaledCursorPosition()
		local x, y, w, h = mapHolder:GetRect()
		local bottomSize = cy >= y and cy <= y + 10
		local topSize = cy <= y + h and cy >= y + h - 10
		local leftSize = cx >= x and cx <= x + 10
		local rightSize = cx <= x + w and cx >= x + w - 10
		
		sizingPoint = nil
		if topSize then
			sizingPoint = "TOP"
		elseif bottomSize then
			sizingPoint = "BOTTOM"
		end
		if leftSize then
			sizingPoint = (sizingPoint or '') .. "LEFT"
		elseif rightSize then
			sizingPoint = (sizingPoint or '') .. "RIGHT"
		end
		if sizingPoint then
			SetCursor([=[Interface\Cursor\Inspect]=])
		else
			SetCursor("POINT_CURSOR")
		end
	end
end

local dragStartX, dragStartY, dragging
function Cartographer3.DragMap(elapsed, currentTime)
	if dragging then
		local cursorX, cursorY = Cartographer3_Utils.GetCursorUniverseCoordinate()
		local diffX, diffY = dragStartX - cursorX, dragStartY - cursorY
		Cartographer3_Data.cameraX = Cartographer3_Data.cameraX + diffX
		Cartographer3_Data.cameraY = Cartographer3_Data.cameraY + diffY
		Cartographer3.Utils.ReadjustCamera()
	end
end

local lastShowMinimapStyle = 0
function Cartographer3.SwitchTextureStyleOnZoom(elapsed, currentTime)
	if Cartographer3_Data.cameraZoom > db.profile.zoomToMinimapTexture then
		if lastShowMinimapStyle == 1 then
			return
		end
		
		Cartographer3_Utils.AddTimer(Cartographer3.ShowMinimapStyleTexturesWhenInsideView)
		
		lastShowMinimapStyle = lastShowMinimapStyle + elapsed/Cartographer3_Data.FADE_TIME
		if lastShowMinimapStyle >= 1 then
			lastShowMinimapStyle = 1
			for zoneTexture, zoneFrame in pairs(zoneFrames) do
				if not Cartographer3_Data.INDOOR_CITIES[zoneTexture] then
					zoneFrame:Hide()
					if zoneFrame:GetAlpha() == 1 then
						zoneFrame:SetAlpha(0.999)
					end
				end
			end
		end
		for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
			minimapStyleTextureFrame:Show()
			minimapStyleTextureFrame:SetAlpha(lastShowMinimapStyle)
		end
	else
		if lastShowMinimapStyle == 0 then
			return
		end
		
		lastShowMinimapStyle = lastShowMinimapStyle - elapsed/Cartographer3_Data.FADE_TIME
		if lastShowMinimapStyle <= 0 then
			lastShowMinimapStyle = 0
			for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
				minimapStyleTextureFrame:Hide()
			end
			Cartographer3_Utils.RemoveTimer(Cartographer3.ShowMinimapStyleTexturesWhenInsideView)
		end
		for _, minimapStyleTextureFrame in pairs(minimapStyleTextureFrames) do
			minimapStyleTextureFrame:SetAlpha(lastShowMinimapStyle)
		end
	end
end

local lastShownTextures = {}
local tmp = {}
function Cartographer3.ShowMinimapStyleTexturesWhenInsideView(elapsed, currentTime)
	lastShownTextures, tmp = tmp, lastShownTextures
	local currentContinentID
	for continentID, currentContinentData in ipairs(Cartographer3_Data.CONTINENT_DATA) do
		local continentLeft = currentContinentData.visibleLeft
		local continentRight = currentContinentData.visibleRight
		local continentBottom = currentContinentData.visibleBottom
		local continentTop = currentContinentData.visibleTop
		if Cartographer3_Data.cameraX >= continentLeft and Cartographer3_Data.cameraX <= continentRight and Cartographer3_Data.cameraY >= continentBottom and Cartographer3_Data.cameraY <= continentTop then
			currentContinentID = continentID
			break
		end 
	end
	if currentContinentID then
		local currentContinentData = Cartographer3_Data.CONTINENT_DATA[currentContinentID]
		local continentTexture = currentContinentData.texture
		local minimapStyleTextureFrame = minimapStyleTextureFrames[continentTexture]
		local center = Cartographer3_Data.cameraX - currentContinentData.mapCenter
		local middle = currentContinentData.mapMiddle - Cartographer3_Data.cameraY
		center = center * Cartographer3_Data.YARDS_PER_PIXEL
		middle = middle * Cartographer3_Data.YARDS_PER_PIXEL
		center = math.floor(center / BLIZZARD_MINIMAP_TILE_YARD_SIZE)
		middle = math.ceil(middle / BLIZZARD_MINIMAP_TILE_YARD_SIZE)
		
		local x_len = math.ceil((scrollFrame:GetWidth()/Cartographer3_Data.cameraZoom * Cartographer3_Data.YARDS_PER_PIXEL / BLIZZARD_MINIMAP_TILE_YARD_SIZE) / 2) + 1
		local y_len = math.ceil((scrollFrame:GetHeight()/Cartographer3_Data.cameraZoom * Cartographer3_Data.YARDS_PER_PIXEL / BLIZZARD_MINIMAP_TILE_YARD_SIZE) / 2) + 1
		for i = center - x_len, center + x_len do
			for j = middle - y_len, middle + y_len do
				local tex = minimapStyleTextureFrame[i * 100 + j + 10000]
				if tex then
					tex:Show()
					lastShownTextures[tex] = true
					tmp[tex] = nil
				end
			end
		end
	end
	for tex in pairs(tmp) do
		tex:Hide()
		tmp[tex] = nil
	end
end

local lastZoneFrame
local currentZoneFrame
local recentShown = {}
function Cartographer3.ShowMapTexturesWhenHovering(elapsed, currentTime)
	local fade_difference = elapsed/Cartographer3_Data.FADE_TIME
	local cursorZone
	if Cartographer3_Utils.IsMouseHovering(scrollFrame) then
		cursorZone = Cartographer3_Utils.GetCursorZonePosition()
	else
		cursorZone = Cartographer3_Data.currentMapTexture
	end
	local isInstance = false
	local zoneFrame = zoneFrames[cursorZone]
	if not zoneFrame then
		isInstance = true
		zoneFrame = instanceTextureFrames[cursorZone]
	end
	if zoneFrame then
		local alpha = zoneFrame:GetAlpha()
		if alpha < 1 or currentZoneFrame ~= zoneFrame then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			recentShown[zoneFrame] = true
			if currentZoneFrame ~= zoneFrame then
				if not currentZoneFrame or not Cartographer3_Data.CITIES[currentZoneFrame.texture] then
					lastZoneFrame = currentZoneFrame
				end
				currentZoneFrame = zoneFrame
			end
			currentZoneFrame:Show()
			currentZoneFrame:SetAlpha(alpha)
		end
	elseif currentZoneFrame then
		local alpha = currentZoneFrame:GetAlpha()
		if alpha < 1 then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			currentZoneFrame:Show()
			currentZoneFrame:SetAlpha(alpha)
		end
	end
	if lastZoneFrame then
		local alpha = lastZoneFrame:GetAlpha()
		if alpha < 1 then
			alpha = alpha + fade_difference
			if alpha >= 1 then
				alpha = 1
			end
			lastZoneFrame:Show()
			lastZoneFrame:SetAlpha(alpha)
		end
	end
	for frame in pairs(recentShown) do
		if frame ~= currentZoneFrame and frame ~= lastZoneFrame then
			local alpha = frame:GetAlpha()
			if alpha > 0 then
				alpha = alpha - fade_difference
				if alpha <= 0 then
					alpha = 0
					frame:Hide()
					recentShown[frame] = nil
				end
				frame:SetAlpha(alpha)
			end
		end
	end
end

local lastToCurrentZone = true
function Cartographer3.ChangeCurrentMapOnZoneHover(elapsed, currentTime)
	if Cartographer3_Utils.IsMouseHovering(scrollFrame) then
		local cursorZone, cursorX, cursorY = Cartographer3_Utils.GetCursorZonePosition()
		if cursorZone == Cartographer3_Data.currentMapTexture then
			return
		end
		local zoneData = Cartographer3_Data.ZONE_DATA[cursorZone]
		if zoneData then
			SetMapZoom(zoneData.continentID, zoneData.id)
			lastToCurrentZone = false
		else
			local continentData = Cartographer3_Data.CONTINENT_DATA[cursorZone]
			if continentData then
				SetMapZoom(continentData.id)
				lastToCurrentZone = false
			else
				SetMapToCurrentZone()
				lastToCurrentZone = true
			end
		end
	else
		if not lastToCurrentZone then
			SetMapToCurrentZone()
			lastToCurrentZone = true
		end
	end
end

local zoneSelectButton, zoomSliderButton

mapHolder = CreateFrame("Frame", "Cartographer_MapHolder", UIParent)
Cartographer3_Data.mapHolder = mapHolder
mapFrame = CreateFrame("Frame", "Cartographer_MapFrame", mapHolder)
Cartographer3_Data.mapFrame = mapFrame
scrollFrame = CreateFrame("ScrollFrame", mapHolder:GetName() .. "_ScrollFrame", mapHolder)
Cartographer3_Data.scrollFrame = scrollFrame
mapView = CreateFrame("Button", mapHolder:GetName() .. "_MapView", scrollFrame)
Cartographer3_Data.mapView = mapView

local function createMapHolder()
	createMapHolder = nil
	mapHolder:SetWidth(db.profile.width)
	mapHolder:SetHeight(db.profile.height)
	mapHolder:SetPoint(db.profile.position[1], UIParent, db.profile.position[1], db.profile.position[2], db.profile.position[3])
	mapHolder:EnableMouse(true)
	mapHolder:SetMovable(true)
	mapHolder:SetResizable(true)
	mapHolder:RegisterForDrag("LeftButton")
	mapHolder:SetMinResize(Cartographer3_Data.MAPFRAME_MINRESIZE_WIDTH, Cartographer3_Data.MAPFRAME_MINRESIZE_HEIGHT)
	mapHolder:SetScript("OnShow", function(this)
		WorldMapFrame:Show()
		db.profile.shown = true
	end)
	mapHolder:SetScript("OnHide", function(this)
		WorldMapFrame:Hide()
		db.profile.shown = false
		SetMapToCurrentZone()
		dragging = false
	end)
	if not db.profile.shown then
		mapHolder:Hide()
	end
	
	mapFrame:SetFrameLevel(mapHolder:GetFrameLevel()+2)
	mapFrame:SetAllPoints(mapHolder)
	
	mapFrame.bg = CreateFrame("Frame", mapFrame:GetName() .. "_Background", mapFrame)
	mapFrame.bg:SetAllPoints(mapFrame)
	mapFrame.bg:SetFrameLevel(mapHolder:GetFrameLevel())
	
	mapFrame.bg:SetBackdrop {
		bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
		edgeFile = [[Interface\DialogFrame\UI-DialogBox-Border]],
		tile = true,
		tileSize = 32,
		edgeSize = 32,
		insets = {
			left = 5,
			right = 6,
			top = 5,
			bottom = 6
		}
	}
	
	positionFontString = mapHolder:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	positionFontString:SetJustifyH("LEFT")
	Cartographer3_Utils.AddTimer(Cartographer3.UpdatePositionFontString)
	Cartographer3_Utils.AddTimer(Cartographer3.FixViewIfMoving)
	Cartographer3_Utils.AddTimer(Cartographer3.ChangeCursorIfOnSide)
	mapHolder:SetScript("OnLeave", function(this)
		mouseDown = false
		SetCursor("POINT_CURSOR")
	end)
	
	mapHolder:SetScript("OnMouseDown", function(this, button)
		if button == "LeftButton" then
			mouseDown = true
		end
	end)
	mapHolder:SetScript("OnMouseUp", function(this, button)
		if button == "LeftButton" then
			mouseDown = false
		end
	end)
	mapHolder:SetScript("OnDragStart", function(this)
		moving = true
		if sizingPoint then
			SetCursor([=[Interface\Cursor\Inspect]=])
			this:StartSizing(sizingPoint)
		else
			SetCursor("POINT_CURSOR")
			this:StartMoving()
		end
	end)
	mapHolder:SetScript("OnDragStop", function(this)
		moving = false
		mouseDown = false
		SetCursor("POINT_CURSOR")
		this:StopMovingOrSizing()
		
		Cartographer3.Utils.ReadjustCamera()
		db.profile.width = this:GetWidth()
		db.profile.height = this:GetHeight()
		
		local width, height = GetScreenWidth(), GetScreenHeight()
		local uiscale = UIParent:GetEffectiveScale()
		local scale = this:GetEffectiveScale() / uiscale
		local x, y = this:GetCenter()
		x, y = x*scale, y*scale
		local point
		if x < width/3 then
			point = "LEFT"
			x = x - this:GetWidth()/2*scale
			if x < -10*scale then
				x = -10*scale
			end
		elseif x < width*2/3 then	
			point = ""
			x = x - width/2
		else	
			point = "RIGHT"
			x = x - width + this:GetWidth()/2*scale
			if x > 10*scale then
				x = 10*scale
			end
		end
		if y < height/3 then
			point = "BOTTOM" .. point
			y = y - this:GetHeight()/2*scale
			if y < -10*scale then
				y = -10*scale
			end
		elseif y < height*2/3 then
			if point == "" then
				point = "CENTER"
			end
			y = y - height/2
		else
			point = "TOP" .. point
			y = y - height + this:GetHeight()/2*scale
			if y > 10*scale then
				y = 10*scale
			end
		end
		x, y = x/scale, y/scale
		this:ClearAllPoints()
		this:SetPoint(point, UIParent, point, x, y)
		db.profile.position[1] = point
		db.profile.position[2] = x
		db.profile.position[3] = y
	end)
	Cartographer3_Utils.AddTimer(Cartographer3.ShowUIOnHover)
	
	scrollFrame:SetScrollChild(mapView)
	scrollFrame:SetPoint("BOTTOMLEFT", mapHolder, "BOTTOMLEFT", 11, 10)
	scrollFrame:SetPoint("TOPRIGHT", mapHolder, "TOPRIGHT", -10, -30)
	mapView:SetWidth(10000)
	mapView:SetHeight(10000)
	Cartographer3.Utils.ReadjustCamera()
	mapView:EnableMouse(true)
	mapView:EnableMouseWheel(true)
	mapView:RegisterForClicks("LeftButtonUp")
	mapView:SetScript("OnUpdate", _G.RequestBattlefieldPositions)
	
	positionFontString:SetPoint("BOTTOMLEFT", scrollFrame, "TOPLEFT", 0, 5)
	
	local bg = mapView:CreateTexture(nil--[[mapView:GetName() .. "_Background"]], "BACKGROUND")
	bg:SetTexture(unpack(Cartographer3_Data.DEFAULT_MAPVIEW_BACKGROUND_COLOR))
	bg:SetAllPoints(mapView)
	createMapTextures()
	mapView:RegisterForDrag("LeftButton")
	Cartographer3_Utils.AddTimer(loadMapOverlayTextureSection, true)
	loadMapOverlayTextureSection()
	Cartographer3_Utils.AddTimer(loadMinimapStyleTextureSection, true)
	loadMinimapStyleTextureSection()
	Cartographer3_Utils.AddTimer(loadInstanceTextureSection, true)
	loadInstanceTextureSection()
	for zoneTexture, zoneFrame in pairs(zoneFrames) do
		if Cartographer3_Data.INDOOR_CITIES[zoneTexture] then
			local p = zoneFrame:GetParent()
			zoneFrame:SetParent(UIParent)
			zoneFrame:SetParent(p)
		end
	end
	Cartographer3_Utils.AddTimer(Cartographer3.DragMap)
	Cartographer3_Utils.AddTimer(Cartographer3.SwitchTextureStyleOnZoom)
	Cartographer3_Utils.AddTimer(Cartographer3.ShowMapTexturesWhenHovering)
	mapView:SetScript("OnMouseDown", function(this, button)
		if button == "LeftButton" then
			dragStartX, dragStartY = Cartographer3_Utils.GetCursorUniverseCoordinate()
		end
	end)
	mapView:SetScript("OnMouseUp", function(this, button)
		if Cartographer3_Utils.IsMouseHovering(_G.WorldMapButton) then
			_G.WorldMapButton:GetScript("OnMouseUp")(_G.WorldMapButton, button)
		end
	end)
	mapView:SetScript("OnDragStart", function(this)
		dragging = true
		Cartographer3_Utils.MoveMap()
	end)
	mapView:SetScript("OnDragStop", function(this)
		dragging = false
	end)
	mapView:SetScript("OnDoubleClick", function(this, arg1)
		local zone = Cartographer3_Utils.GetCursorZonePosition()
		if zone then
			Cartographer3_Utils.ZoomToZone(zone)
		end
	end)
	
	mapView:SetScript("OnMouseWheel", function(this, arg1)
		local cursorX, cursorY = Cartographer3_Utils.GetCursorUniverseCoordinate()
		local zoomIn = arg1 > 0
		local zoom = Cartographer3_Data.gradualCameraZoom or Cartographer3_Data.cameraZoom
		if zoomIn then
			zoom = zoom / Cartographer3_Data.ZOOM_STEP
			if zoom > Cartographer3_Data.MAXIMUM_ZOOM then
				zoom = Cartographer3_Data.MAXIMUM_ZOOM
			end
		else
			zoom = zoom * Cartographer3_Data.ZOOM_STEP
			if zoom < Cartographer3_Data.MINIMUM_ZOOM then
				zoom = Cartographer3_Data.MINIMUM_ZOOM
			end
		end
		if zoom == Cartographer3_Data.cameraZoom then
			Cartographer3_Utils.MoveMap()
		else
			if Cartographer3_Utils.GetMoveMapMessage() == 'player' then
				Cartographer3_Utils.MoveMap(
					nil,
					nil,
				    zoom,
					true,
					'player'
				)
			else
				local cursorX, cursorY = Cartographer3_Utils.GetCursorUniverseCoordinate()
			
				Cartographer3_Utils.MoveMap(
					(Cartographer3_Data.cameraX - cursorX) * Cartographer3_Data.cameraZoom / zoom + cursorX,
					(Cartographer3_Data.cameraY - cursorY) * Cartographer3_Data.cameraZoom / zoom + cursorY,
				    zoom,
					true
				)
			end
		end
	end)
	
	Cartographer3_Utils.AddTimer(Cartographer3.ChangeCurrentMapOnZoneHover)
	
	Cartographer3_Utils.AddTimer(Cartographer3.UpdatePOITooltips)
	
	local closeButton = CreateFrame("Button", mapFrame:GetName() .. "_CloseButton", mapFrame, "UIPanelCloseButton")
	closeButton:SetPoint("TOPRIGHT", -7, -9)
	closeButton:SetScript("OnClick", function(this)
		mapHolder:Hide()
	end)
	closeButton:GetNormalTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:GetPushedTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:GetHighlightTexture():SetTexCoord(0.17, 0.8, 0.2, 0.8)
	closeButton:SetWidth(22)
	closeButton:SetHeight(22)
	
	positionFontString:SetPoint("RIGHT", closeButton, "LEFT", -3, 0)
	
	local button_num = 0
	local lastLeft, lastRight
	local function makeButton(text, side, clickFunc, check, initialChecked)
		button_num = button_num + 1
		local button = CreateFrame(check and "CheckButton" or "Button", mapFrame:GetName() .. "_Button" .. button_num, mapFrame)
		if side == "RIGHT" then
			if not lastRight then
				button:SetPoint("BOTTOMRIGHT", mapFrame, "BOTTOMRIGHT", -7, 9)
			else
				button:SetPoint("RIGHT", lastRight, "LEFT", -3, 0)
			end
			lastRight = button
		else
			if not lastLeft then
				button:SetPoint("BOTTOMLEFT", mapFrame, "BOTTOMLEFT", 8, 9)
			else
				button:SetPoint("LEFT", lastLeft, "RIGHT", 3, 0)
			end
			lastLeft = button
		end
		button:SetWidth(22)
		button:SetHeight(22)
		button:EnableMouse(true)
		button:SetText(text)
		button:RegisterForClicks("LeftButtonUp")
		
		local left = button:CreateTexture(nil--[[button:GetName() .. "Left"]], "BACKGROUND")
		left:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		left:SetWidth(12)
		left:SetHeight(22)
		left:SetTexCoord(0, 0.09375, 0, 0.6875)
		left:SetPoint("LEFT")
		local right = button:CreateTexture(nil--[[button:GetName() .. "Right"]], "BACKGROUND")
		right:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		right:SetWidth(12)
		right:SetHeight(22)
		right:SetTexCoord(0.53125, 0.625, 0, 0.6875)
		right:SetPoint("RIGHT")
		local middle = button:CreateTexture(nil--[[button:GetName() .. "Middle"]], "BACKGROUND")
		middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		middle:SetTexCoord(0.09375, 0.53125, 0, 0.6875)
		middle:SetPoint("LEFT", left, "RIGHT")
		middle:SetPoint("RIGHT", right, "LEFT")
		
		local highlight = button:CreateTexture(nil--[[button:GetName() .. "Highlight"]], "HIGHLIGHT")
		highlight:SetTexture([[Interface\Buttons\UI-Panel-Button-Highlight]])
		highlight:SetTexCoord(0, 0.625, 0, 0.6875)
		highlight:SetBlendMode("ADD")
		highlight:SetAllPoints()

		button:SetHighlightTexture(highlight)
		local function setDown(this)
			button:GetFontString():SetPoint("CENTER", 1.28, -1.28)
			left:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
			right:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
			middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Down]])
		end
		local function setUp(this)
			button:GetFontString():SetPoint("CENTER", 0, 0)
			left:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
			right:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
			middle:SetTexture([[Interface\Buttons\UI-Panel-Button-Up]])
		end
		button:SetScript("OnMouseDown", function(this)
			local checked = this.GetChecked and this:GetChecked()
			if checked then
				setUp(this)
			else
				setDown(this)
			end
		end)
		button:SetScript("OnMouseUp", function(this)
			local checked = this.GetChecked and this:GetChecked()
			if checked then
				setDown(this)
			else
				setUp(this)
			end
		end)
		button:SetScript("OnClick", function(this, button)
			if this.GetChecked then
				clickFunc(not not this:GetChecked())
				this:GetScript("OnMouseUp")(this)
			else
				clickFunc()
			end
		end)
		button:SetTextFontObject(GameFontNormal)
		button:SetText(text)
		button:SetPushedTextOffset(0, 0)
		if check and initialChecked then
			button:SetChecked(true)
			setDown(button)
		end
		return button
	end
	local openConfigButton = makeButton("C", "RIGHT", function()
		InterfaceOptionsFrame_OpenToFrame("Cartographer3")
	end, false)
	local followPlayer = db.profile.followPlayer
	local followPlayerButton = makeButton("P", "RIGHT", Cartographer3.SetFollowPlayer, true, followPlayer)
	Cartographer3.SetFollowPlayer(followPlayer)
	local zoomToZoneButton = makeButton("Z", "RIGHT", Cartographer3_Utils.ZoomToCurrentZone, false)
	zoneSelectButton = makeButton("G", "LEFT", Cartographer3.OpenGotoDropdownMenu, false)
	zoomSliderButton = makeButton("S", "RIGHT", Cartographer3.OpenZoomSlider, false)
	
	WorldMapTooltip:SetParent(mapHolder)
	WorldMapTooltip:SetFrameStrata("TOOLTIP")
	WorldMapFrame:SetParent(mapView)
	WorldMapFrame:SetAlpha(0)
	WorldMapFrame:SetScale(1e-5)
	if mapHolder:IsShown() then
		WorldMapFrame:Show()
	end
	_G.WorldMapDetailFrame:SetParent(mapView)
	_G.WorldMapButton:SetParent(mapView)
	_G.WorldMapPositioningGuide:SetParent(mapView)
	_G.WorldMapDetailFrame:Show()
	_G.WorldMapButton:Show()
	_G.WorldMapPositioningGuide:Show()
	
--	if UnitIsDeadOrGhost("player") then
--		Cartographer3:PLAYER_DEAD()
--	end
end

local dropdownFrame
local gotoDropdownMenu

function Cartographer3.OpenGotoDropdownMenu()
	if not dropdownFrame then
		dropdownFrame = CreateFrame("Frame", mapHolder:GetName() .. "_Dropdown", mapHolder, "UIDropDownMenuTemplate")
	end
	if not gotoDropdownMenu then
		local function menuSort(alpha, bravo)
			return alpha.text < bravo.text
		end
		gotoDropdownMenu = {}
		for i, v in ipairs(Cartographer3_Data.CONTINENT_DATA) do
			local t = {
				text = v.localizedName,
				hasArrow = true,
				func = Cartographer3_Utils.ZoomToZone,
				arg1 = v.texture,
				menuList = {}
			}
			local menuList = t.menuList
			local cities, zones = {}, {}
			for texture, zone in pairs(v.zones) do
				local x = {
					text = zone.localizedName,
					func = Cartographer3_Utils.ZoomToZone,
					arg1 = texture,
				}
				if Cartographer3_Data.CITIES[texture] then
					cities[#cities+1] = x
				else
					zones[#zones+1] = x
				end
			end
			table.sort(cities, menuSort)
			table.sort(zones, menuSort)
			menuList[#menuList+1] = {
				text = L["Cities"],
				isTitle = true,
			}
			for i,v in ipairs(cities) do
				menuList[#menuList+1] = v
			end
			menuList[#menuList+1] = {
				text = L["Zones"],
				isTitle = true,
			}
			for i,v in ipairs(zones) do
				menuList[#menuList+1] = v
			end
			gotoDropdownMenu[i] = t
		end
		table.sort(gotoDropdownMenu, menuSort)
	end
	if DropDownList1:IsShown() and dropdownFrame.menuList == gotoDropdownMenu then
		CloseDropDownMenus()
	else
		EasyMenu(gotoDropdownMenu, dropdownFrame, zoneSelectButton, 0, 0, "MENU")
	end
end

local sliderFrame
function Cartographer3.OpenZoomSlider()
	if not sliderFrame then
		sliderFrame = CreateFrame("Frame", mapFrame:GetName() .. "_SliderFrame", mapFrame)
		Cartographer3_Data.sliderFrame = sliderFrame
		sliderFrame:Hide()
		
		sliderFrame:SetWidth(46)
		sliderFrame:SetHeight(170)
		sliderFrame:SetBackdrop({
			bgFile = [[Interface\Tooltips\UI-Tooltip-Background]],
			edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
			tile = true,
			insets = {
				left = 5,
				right = 5,
				top = 5,
				bottom = 5
			},
			tileSize = 16,
			edgeSize = 16
		})
		sliderFrame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
		sliderFrame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
		sliderFrame:EnableMouse(true)
		sliderFrame:Hide()
		sliderFrame:SetPoint("CENTER", UIParent, "CENTER")
		local slider = CreateFrame("Slider", sliderFrame:GetName() .. "_Slider", sliderFrame)
		sliderFrame.slider = slider
		slider:SetOrientation("VERTICAL")
		slider:SetMinMaxValues(0, 1)
		slider:SetValueStep(1e-10)
		slider:SetValue(0.5)
		slider:SetWidth(16)
		slider:SetHeight(128)
		slider:SetPoint("CENTER", sliderFrame, "CENTER", 0, 0)
		slider:SetBackdrop({
			bgFile = [[Interface\Buttons\UI-SliderBar-Background]],
			edgeFile = [[Interface\Buttons\UI-SliderBar-Border]],
			tile = true,
			edgeSize = 8,
			tileSize = 8,
			insets = {
				left = 3,
				right = 3,
				top = 3,
				bottom = 3
			}
		})
		slider:SetThumbTexture([[Interface\Buttons\UI-SliderBar-Button-Vertical]])
		
		local MAXIMUM_ZOOM_log = math.log(Cartographer3_Data.MAXIMUM_ZOOM)
		local MINIMUM_ZOOM_log = math.log(Cartographer3_Data.MINIMUM_ZOOM)
		local changing = false
		slider:SetScript("OnValueChanged", function(this)
			if changing then
				return
			end
			local value = (1 - this:GetValue()) * (MAXIMUM_ZOOM_log - MINIMUM_ZOOM_log) + MINIMUM_ZOOM_log
			Cartographer3_Utils.MoveMap(nil, nil, math.exp(value))
		end)
		local closeTime = 0
		sliderFrame:SetScript("OnShow", function(this)
			closeTime = GetTime() + 5
		end)
		sliderFrame:SetScript("OnUpdate", function(this)
			local currentTime = GetTime()
			if Cartographer3_Utils.IsMouseHovering(this) then
				closeTime = currentTime + 3
			end
			if currentTime > closeTime then
				this:Hide()
			end
		end)
		
		function slider:set(value)
			changing = true
			
			self:SetValue(-(value - MINIMUM_ZOOM_log)/(MAXIMUM_ZOOM_log - MINIMUM_ZOOM_log) + 1)
			
			changing = false
		end
	end
	if sliderFrame:IsShown() then
		sliderFrame:Hide()
	else
		sliderFrame:Show()
		sliderFrame:SetPoint("BOTTOMRIGHT", zoomSliderButton, "TOPRIGHT")
	end
end

local lastPlayerX, lastPlayerY = 0, 0
function Cartographer3.FollowPlayerIfMoved()
	if Cartographer3_Utils.GetMoveMapMessage() ~= 'player' and Cartographer3_Utils.IsMouseHovering(mapHolder) then
		return
	end
	local playerX, playerY = Cartographer3_Utils.GetUnitUniverseCoordinate("player")
	if not playerX or math.abs(lastPlayerX - playerX) < 1e-10 and math.abs(lastPlayerY - playerY) < 1e-10 then
		return
	end
	lastPlayerX, lastPlayerY = playerX, playerY
	Cartographer3_Utils.ZoomToPlayer()
end

function Cartographer3.SetFollowPlayer(follow)
	db.profile.followPlayer = follow
	if follow then
		lastPlayerX, lastPlayerY = 0, 0
		Cartographer3_Utils.AddTimer(Cartographer3.FollowPlayerIfMoved)
		Cartographer3_Utils.ZoomToPlayer()
	else
		if Cartographer3_Utils.GetMoveMapMessage() == 'player' then
			Cartographer3_Utils.MoveMap()
		end
		Cartographer3_Utils.RemoveTimer(Cartographer3.FollowPlayerIfMoved)
	end
end

function Cartographer3.SetFrameStrata(value)
	mapHolder:SetFrameStrata(value)
	Cartographer3.db.profile.strata = value
end

local function ADDON_LOADED(event)
	Cartographer3_Utils.RemoveEventListener("ADDON_LOADED", ADDON_LOADED)
	ADDON_LOADED = nil
	db = LibStub("AceDB-3.0"):New("Cartographer3DB", {
		profile = {
			width = Cartographer3_Data.DEFAULT_MAPFRAME_WIDTH,
			height = Cartographer3_Data.DEFAULT_MAPFRAME_HEIGHT,
			position = { "CENTER", 0, 0 },
			shown = true,
			followPlayer = true,
			zoomToMinimapTexture = Cartographer3_Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE,
			strata = "FULLSCREEN",
		}
	}, 'global')
	Cartographer3.db = db
	Cartographer3.SetFrameStrata(db.profile.strata)
	local Cartographer3_NewOverlaysDB = _G.Cartographer3_NewOverlaysDB
	if type(Cartographer3_NewOverlaysDB) == "table" then
		for zoneTexture, overlayData in pairs(Cartographer3_NewOverlaysDB) do
			if type(overlayData) ~= "table" then
				Cartographer3_NewOverlaysDB[zoneTexture] = {}
			end
			if not Cartographer3_Data.OVERLAY_DATA[zoneTexture] then
				Cartographer3_Data.OVERLAY_DATA[zoneTexture] = {}
			end
			for tname, num in pairs(overlayData) do
				if Cartographer3_Data.OVERLAY_DATA[zoneTexture][tname] == num then
					overlayData[tname] = nil -- it's been upgraded in the main addon
				else
					Cartographer3_Data.OVERLAY_DATA[zoneTexture][tname] = num
				end
			end
			if not next(overlayData) then
				Cartographer3_NewOverlaysDB[zoneTexture] = nil
			end
		end
		if not next(Cartographer3_NewOverlaysDB) then
			_G.Cartographer3_NewOverlaysDB = nil
		end
	end
	
	initializeContinentData()
	createMapHolder()
end
Cartographer3_Utils.AddEventListener("ADDON_LOADED", ADDON_LOADED)

local function UPDATE_PENDING_MAIL(event)
	Cartographer3_Utils.RemoveEventListener("UPDATE_PENDING_MAIL", UPDATE_PENDING_MAIL)
	UPDATE_PENDING_MAIL = nil
	local timeToRun = GetTime() + 1
	local function func(elapsed, currentTime)
		if currentTime < timeToRun then
			return
		end
		Cartographer3_Utils.RemoveTimer(func)
		func = nil
		Cartographer3_Utils.ZoomToCurrentZone()
	end
	Cartographer3_Utils.AddTimer(func)
end
Cartographer3_Utils.AddEventListener("UPDATE_PENDING_MAIL", UPDATE_PENDING_MAIL)

function _G.ToggleWorldMap()
	if mapHolder:IsShown() then
		mapHolder:Hide()
	else
		mapHolder:Show()
	end
end